using System;
using HarmonyLib;
using Vintagestory.API.Client;
using Vintagestory.API.Common;
using Vintagestory.API.Common.Entities;
using Vintagestory.API.MathTools;
using Vintagestory.GameContent;

namespace AltMapIconRenderer {
	public class AltEntityMapComponent {
		public static bool Render(EntityMapComponent __instance, GuiElementMap map, float dt) {
			Traverse self = Traverse.Create(__instance);
			ICoreClientAPI capi = self.Field("capi").GetValue() as ICoreClientAPI;
			Entity entity = self.Field("entity").GetValue() as Entity;
			MeshRef quadModel = self.Field("quadModel").GetValue() as MeshRef;
			LoadedTexture Texture = self.Field("Texture").GetValue() as LoadedTexture;
			Vec2f viewPos = self.Field("viewPos").GetValue() as Vec2f;
			Matrixf mvMat = self.Field("mvMat").GetValue() as Matrixf;

			var player = (entity as EntityPlayer)?.Player;
			if(player?.WorldData?.CurrentGameMode == EnumGameMode.Spectator == true) {
				return false;
			}
			if((entity as EntityPlayer)?.Controls.Sneak == true && player != capi.World.Player) {
				return false;
			}

			map.TranslateWorldPosToViewPos(entity.Pos.XYZ, ref viewPos);

			float x = (float)(map.Bounds.renderX + viewPos.X);
			float y = (float)(map.Bounds.renderY + viewPos.Y);
			
			ICoreClientAPI api = map.Api;

			if(Texture.Disposed) {
				throw new Exception("Fatal. Trying to render a disposed texture");
			}
			if(quadModel.Disposed) {
				throw new Exception("Fatal. Trying to render a disposed texture");
			}

			capi.Render.GlToggleBlend(true);

			IShaderProgram prog = api.Render.GetEngineShader(EnumShaderProgram.Gui);
			prog.Uniform("rgbaIn", AltMapIconRendererSystem.playerColour);
			prog.Uniform("extraGlow", 0);
			prog.Uniform("applyColor", 0);
			prog.Uniform("noTexture", 0f);
			prog.BindTexture2D("tex2d", Texture.TextureId, 0);

			mvMat.Set(api.Render.CurrentModelviewMatrix)
					.Translate(x, y, 60)
					.Scale(Texture.Width, Texture.Height, 0)
					.Scale(AltMapIconRendererSystem.playerScale, AltMapIconRendererSystem.playerScale, 0)
					.RotateZ(-entity.Pos.Yaw + 90 * GameMath.DEG2RAD);

			prog.UniformMatrix("projectionMatrix", api.Render.CurrentProjectionMatrix);
			prog.UniformMatrix("modelViewMatrix", mvMat.Values);

			api.Render.RenderMesh(quadModel);

			// Skip original method.
			return false;
		}
	}
}